/*{{{  filer.c*/
#include <stdio.h>
#define STDIO_H
#include <ctype.h>
#include "adaptdat.h"
#include "lex.h"
#include "strmacs.h"
#include "symbol.h"
#include "typedef.h"

/*{{{  #defines*/
#ifdef  V_ADAPT

#define  AD_PROC_ARRAY	0
#define  AD_TEMPSIZE	1
#define  N_ADAPT_KEYWORDS   (sizeof (adapt_keyword) / sizeof (Adaptin))

#endif
/*}}}*/
/*{{{  typedefs: Adaptin, Filerec*/
#ifdef  V_ADAPT

typedef struct Adaptin {
	char *name;
	bool  defined;
} Adaptin;

#endif

typedef struct Filerec {
	char *filename;
	FILE *ifp;
	int   lineno;
	char  next_label [8];	/* only used for 'fixed' source form */
	int   start_col;	/*   "    "   "     "      "      "  */    
	bool  need_label;	/*   "    "   "     "      "      "  */    
	struct Filerec *prev;
} Filerec;

/*}}}*/
/*{{{  globals*/
#ifdef  V_ADAPT

int  tempsize;
int  proc_rank, proc_size [MAX_RANK+2];
	/*------------------------------------------------------*/
	/* +2, since it starts at 'proc_size [1]', and if more 	*/
	/* than MAX_RANK dims are declared, then		*/
	/* 'proc_size [MAX_RANK+1]' will be used!		*/
	/*------------------------------------------------------*/

Adaptin  adapt_keyword [] = {
	"proc_array",	FALSE,
	"tempsize",	FALSE
};

#endif

Filerec  *last_filerecp = NULL;
/*}}}*/

#ifdef  V_ADAPT
/*---------------------*/
/* reading 'adapt.dat' */
/*---------------------*/
/*{{{  read_adapt_dat (fp) -- not needed by IDA*/
read_adapt_dat (fp)
/*---------------*/
	FILE *fp;
/*----------------------------------------------------------------------*/
/*	This expects to find declarations in the following form:	*/
/*		proc_array   (1,2,3)					*/
/*		tempsize     200000					*/
/* The list of dimensions in 'proc_array' can be empty.			*/
/* 'tempsize' is the size of the 'stack vector' in integer words.	*/
/*----------------------------------------------------------------------*/
{
	/*{{{  decls*/
	bool  found_bracket = FALSE;
	char *ptr,  *word;
	int   c,  i,  lineno = 0;
	/*}}}*/

	while (ptr = fgets (textbuf, TEXTBUF_LEN-1, fp)) {
		lineno++;
		get_nonwhite_from_here (c, ptr);
		if (c == '!' || c == '\n') continue;

		if (! isalpha (c)) {
			/*{{{  error - line must start with a 'keyword'*/
			fprintf (stderr, "\"adapt.dat\" line %d: line must start with a \'keyword\'\n",
				lineno);
			exit (1);
			/*}}}*/
		}
		/*{{{  read the word, and flag error if not a 'keyword'*/
		
		word = ptr;
		for (c = *++ptr;   (isalpha (c) || c == '_');  c = *++ptr)
			;
		
		*ptr = '\0';	/* the char at this position is stored in 'c' */
		
		for (i = 0;  i < N_ADAPT_KEYWORDS;  i++)
			if (str_match (adapt_keyword [i].name,  word)) {	/* found! */
				adapt_keyword [i].defined = TRUE;
				break;
			}
		
		if (i == N_ADAPT_KEYWORDS) {
			fprintf (stderr, "\"adapt.dat\" line %d: \'%s\' is not a \'keyword\'\n",
				lineno,  word);
			exit (1);
		}
		
		*ptr = c;
		/*}}}*/

		switch (i) {
		case  AD_PROC_ARRAY :
			/*{{{  read proc array size*/
			
			get_nonwhite_from_here (c, ptr);
			if (c != '(') goto syntax_err;
			
			
			i = 0;
			proc_size [0] = 1;	/* so can use the same formulae for	*/
						/* both distributed and internal dims.	*/
			
			ptr++;			/* to point to char after '(', for 'read ()' fns */
				
			if (filer_read_num (&ptr, &proc_size[++i])) {
				while (read_comma_num (&ptr, &proc_size[++i])) {
					if (i > MAX_RANK) {
						fprintf (stderr, "\"adapt.dat\": process array has more than %d dimensions\n", MAX_RANK);
						exit (1);
					}
				}
			}
			proc_rank = i-1;
			
			get_nonwhite_from_here (c, ptr);
			if (c != ')')  goto syntax_err;
			
			for (i=1; i <= proc_rank; i++) {
				if (proc_size[i] <= 0) {
					fprintf (stderr, "\"adapt.dat\": process array size must be >0 in all dimensions\n");
					exit (1);
				}
			}
			break;
			/*}}}*/

		case  AD_TEMPSIZE :
			/*{{{  read 'stack vector' size*/
			
			if ( ! filer_read_num (&ptr, &tempsize)) {
				fprintf (stderr, "\"adapt.dat\" line %d: missing value for \'tempsize\'\n",
					lineno);
				exit (1);
			}
			if (tempsize < 0) {
				fprintf (stderr, "\"adapt.dat\": \'tempsize\' value must be positive\n");
				exit (1);
			}
			break;
			
			/*}}}*/
		}
	}

	/*{{{  check that all required info has been supplied*/
	for (i = 0;  i < N_ADAPT_KEYWORDS;  i++)
		if ( ! adapt_keyword [i].defined) {
			fprintf (stderr, "\"adapt.dat\": \'%s\' has not been specified\n",
				adapt_keyword[i].name);
			exit (1);
		}
	/*}}}*/

	fclose (fp);
	/*{{{  write 'process array size' at start of o/p program*/
	
	printf ("C\n");
	if (proc_rank > 0) {
		printf ("C Program generated by ADAPT for a process array of size (");
		printf ("%d", proc_size[1]);
		for (i=2; i <= proc_rank; i++)
			printf (",%d", proc_size[i]);
		printf (").\nC\n");
	}
	else
		printf ("C This program was generated by ADAPT for a single process\nC\n");
	
	/*}}}*/
	
	return;

syntax_err:
	fprintf (stderr, "\"adapt.dat\" line %d: syntax error\n", lineno);
	exit (1);
}
/*}}}*/
/*{{{  bool filer_read_num (pptr, pnum)*/
bool filer_read_num (pptr, pnum)
/*----------------------------*/
	char **pptr;
	int   *pnum;
/*--------------------------------------------------------------*/
/*	Scans 'textbuf', starting at '*pptr', for the next	*/
/* nonblank character 'c'.  If it's a digit or '-' followed by	*/
/* a digit, the number is read up to the next non-digit, its	*/
/* value is stored in '*pnum', '*pptr' is left pointing to the	*/
/* next char, and the function returns TRUE.  Otherwise,	*/
/* '*pptr' is left pointing to the next unrecognised char	*/
/* and FALSE is returned.					*/ 
/*--------------------------------------------------------------*/
{
	register char  *ptr = *pptr;
	bool  pos = TRUE;
	int   c,  val;
	
	get_nonwhite_from_here (c, ptr);
	if (c == '-') {
		pos = FALSE;
		c = *++ptr;
	}
	if (isdigit (c)) {
		val = c - '\0';
		for (c = *++ptr;   isdigit (c);   c = *++ptr)
			val = 10 * val + (c - '0');

		*pnum = pos ? val : -val;
		*pptr = ptr;
		return TRUE;
	}
	else {
		*pptr = ptr;
		return FALSE;
	}
}
/*}}}*/
/*{{{  bool read_comma_num (pptr, pnum)*/
bool read_comma_num (pptr, pnum)
/*----------------------------*/
	char **pptr;
	int   *pnum;
/*--------------------------------------------------------------*/
/*	Scans 'textbuf', starting at '*pptr'.  If the next	*/
/* nonwhite fields in the string are a comma followed by an	*/
/* integer (possibly preceded by a '-' sign, and with optional	*/
/* space before and after the comma), the integer value is put	*/
/* in '*pnum', '*pptr' is left pointing to the next char, and	*/
/* the function returns TRUE.  Otherwise, '*pptr' is left	*/
/* pointing to the next nonwhite char, and FALSE is returned.	*/
/*--------------------------------------------------------------*/
{
	register char  *ptr = *pptr;
	bool   pos = TRUE;
	char  *next_char;
	int    c,  val;
	
	get_nonwhite_from_here (c, ptr);
	next_char = ptr;
	if (c != ',') {
		*pptr = next_char;
		return FALSE;
	}

	get_next_nonwhite (c, ptr);
	if (c == '-') {
		pos = FALSE;
		c = *++ptr;
	}
	if (isdigit (c)) {
		val = c - '\0';
		for (c = *++ptr;   isdigit (c);   c = *++ptr)
			val = 10 * val + (c - '0');

		*pnum = pos ? val : -val;
		*pptr = ptr;
		return TRUE;
	}
	else {
		*pptr = next_char;
		return FALSE;
	}
}
/*}}}*/

#endif

/*------------------------------*/
/* input & output file handling */
/*------------------------------*/
/*{{{  bool open_include_file (incl_filename)*/
open_include_file (incl_filename)
/*-----------------------------*/
	char *incl_filename;
{
	char  c,  *from,  *to;
	FILE  *incl_fp;
	Filerec  *filerecp;

	/*{{{  strip the delimiters (' or ") from 'incl_filename'*/
	/*------------------------------------------------------*/
	/* strip the delimiters (' or ") from "incl_filename".	*/
	/*------------------------------------------------------*/
	to   = incl_filename - 1;
	from = incl_filename;
	while (*++to = *++from)
		;
	c = *--to;
	if (c == '\'' || c == '\"') *to = '\0';
	
	/*}}}*/
	
	/*{{{  check that ref to 'incl_filename' is not recursive*/
	/*--------------------------------------------------------------*/
	/* check that reference to 'incl_filename' is not recursive ---	*/
	/* i.e. that the file is not already open!			*/
	/*--------------------------------------------------------------*/
	
	if (src_filename) {
		/*------------------------------------------------------*/
		/* else input is currently from 'stdin' so this check	*/
		/* is unnecessary.					*/
		/*------------------------------------------------------*/
		if (str_match (incl_filename, src_filename)) {
			errmsg ("a file must not INCLUDE itself");
			return;
		}
		for (filerecp = last_filerecp;  filerecp;  filerecp = filerecp->prev) {
			if (str_match (incl_filename, filerecp->filename)) {
				/*{{{  indirect recursive INCLUDE: give errmsg showing chain*/
				/*--------------------------------------------------------------*/
				/* Recursive INCLUDE via intermediate files.  Don't use the	*/
				/* normal 'errmsg()' function, as we want the errmsg to start	*/
				/* with 'incl_filename' rather than the current 'filename'.	*/
				/*--------------------------------------------------------------*/
				fprintf (stderr, "\"%s\": this file illegally INCLUDEs itself via the INCLUDE chain \'%s\' -> ",
						incl_filename, incl_filename);
				
				/*--------------------------------------------------------------*/
				/* Output the name of each file in the 'file stack'...		*/
				/*--------------------------------------------------------------*/
				while (filerecp != last_filerecp) {
					/*------------------------------------------------------*/
					/* Find the next file after '*filerecp' ('*frec') --	*/
					/* we know it's not NULL!				*/
					/*------------------------------------------------------*/
					Filerec  *frecp;
				
					for (frecp = last_filerecp;  frecp->prev != filerecp;  frecp = frecp->prev)
						;
				
					fprintf (stderr, "\'%s\' -> ", frecp->filename);
					filerecp = frecp;
				}
				
				fprintf (stderr, "\'%s\' -> \'%s\'\n", src_filename, incl_filename);
				
				return;
				/*}}}*/
			}
		}
	}
	/*}}}*/
	
	/*{{{  open file "incl_filename" -- if can't open it, return*/
	/*------------------------------*/
	/* open file "incl_filename"	*/
	/*------------------------------*/
	if ((incl_fp = fopen (incl_filename, "r")) == NULL) {
		errmsg ("can't open \'INCLUDE\' file %q", incl_filename);
		return;
	}
	/*}}}*/
	
	/*--------------------------------------------------------------*/
	/* save info about current input file in an 'Filerec' record.	*/
	/*--------------------------------------------------------------*/
	filerecp = (Filerec *) emalloc (sizeof (Filerec));
	
	filerecp->filename = src_filename;	/* current filename -- global */
	filerecp->ifp      = ifp;
	filerecp->lineno   = lineno;
	if ( !free_source_form) {
		strcpy (filerecp->next_label, next_label);
		filerecp->start_col   = start_col;
		filerecp->need_label = need_label;
	}
	filerecp->prev = last_filerecp;	
	last_filerecp  = filerecp;

	/*----------------------------------------------*/
	/* update global file info for new input file.	*/
	/*----------------------------------------------*/
	src_filename = incl_filename;
	ifp          = incl_fp;
	lineno       = 0;
	need_label   = TRUE;	/* used for 'fixed' source form */
}
/*}}}*/
/*{{{  FILE *close_file ()*/
FILE *close_file ()
/*---------------*/
{
	FILE *new_fp;
	Filerec  *filerecp;

	fclose (ifp);

	if (!last_filerecp)
		return  NULL;

	src_filename = last_filerecp->filename;
	new_fp       = last_filerecp->ifp;
	lineno       = last_filerecp->lineno;
	if ( !free_source_form) {
		strcpy (next_label, last_filerecp->next_label);
		start_col   = last_filerecp->start_col;
		need_label = last_filerecp->need_label;
	}

	/*------------------------------------------------------*/
	/* free the current 'last_filerecp' record, and reset	*/
	/* the 'last_filerecp' pointer.				*/
	/*------------------------------------------------------*/
	filerecp = last_filerecp->prev;
	free ((char *) last_filerecp);
	last_filerecp = filerecp;
	
	return  new_fp;
}
/*}}}*/
/*}}}*/
